module Parse (parseString, parseExpression, parseFile) where

import System.IO
import Control.Monad
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Expr
import Text.ParserCombinators.Parsec.Language
import qualified Text.ParserCombinators.Parsec.Token as Token

import Types

language =
  emptyDef { Token.commentStart    = "/*",
             Token.commentEnd      = "*/",
             Token.commentLine     = "//",
             Token.identStart      = letter,
             Token.identLetter     = alphaNum,
             Token.reservedNames   = [ "if", "then", "else", "while", "do",
                                       "and", "or", "not", "true", "false", "print"],
             Token.reservedOpNames = ["+", "-", "*", "/", ":=", "==", "<", ">", "and", "or", "not"]
           }

identifier = Token.identifier lexer
reserved   = Token.reserved   lexer
reservedOp = Token.reservedOp lexer
parens     = Token.parens     lexer
braces     = Token.braces     lexer
integer    = Token.integer    lexer
semi       = Token.semi       lexer
whiteSpace = Token.whiteSpace lexer
commaSep   = Token.commaSep   lexer

lexer = Token.makeTokenParser language

parser = whiteSpace >> stmtSequence >>= \s -> eof >> return s

stmtSequence =
    do ss <- many1 atomicStmt
       return $ case ss of
                  [s] -> s
                  _ -> Block ss

atomicStmt = braces stmtSequence <|> ifStmt <|> whileStmt <|> printStmt <|> assignStmt <?> "statement"

ifStmt =
    do reserved "if"
       cond <- expression
       reserved "then"
       stmt1 <- atomicStmt
       stmt2 <- (reserved "else" >> atomicStmt) <|> return Empty
       return $ If cond stmt1 stmt2

whileStmt =
    do reserved "while"
       cond <- expression
       reserved "do"
       stmt <- atomicStmt
       return $ While cond stmt

printStmt =
    do reserved "print"
       e <- expression
       semi
       return $ Print e

assignStmt =
    do name <- identifier
       reservedOp ":="
       expr <- expression
       semi
       return $ Assign name expr

expression = buildExpressionParser operators term

operators = [[Prefix (reservedOp "-" >> return (UnaryOp "-"))],
             [Infix (reservedOp "*" >> return (BinaryOp "*")) AssocLeft,
              Infix (reservedOp "/" >> return (BinaryOp "/")) AssocLeft],
             [Infix (reservedOp "+" >> return (BinaryOp "+")) AssocLeft,
              Infix (reservedOp "-" >> return (BinaryOp "-")) AssocLeft],
             [Infix (reservedOp "==" >> return (BinaryOp "==")) AssocLeft,
              Infix (reservedOp "<" >> return (BinaryOp "<")) AssocLeft,
              Infix (reservedOp ">" >> return (BinaryOp ">")) AssocLeft],
             [Infix (reservedOp "and" >> return (BinaryOp "and")) AssocLeft,
              Infix (reservedOp "or" >> return (BinaryOp "or")) AssocLeft,
              Prefix (reservedOp "not" >> return (UnaryOp "not"))]]
              
term = parens expression <|> boolLitTerm <|> liftM Var identifier <|> liftM IntLit integer

boolLitTerm = (reserved "true" >> return (BoolLit True)) <|> (reserved "false" >> return (BoolLit False))

parseString :: String -> Statement
parseString str =
  case parse parser "" str of
      Left e  -> error $ show e
      Right r -> r

parseExpression :: String -> Expression
parseExpression str =
  case parse expression "" str of
      Left e  -> error $ show e
      Right r -> r

parseFile :: String -> IO Statement
parseFile file =
    do prog <- readFile file
       return $ parseString prog
